package Geo::Coder::Free::DB::MaxMind::admin1;

use strict;
use warnings;

# admin1.db is from http://download.geonames.org/export/dump/admin1CodesASCII.txt

use Geo::Coder::Free::DB::MaxMind::admin;

our @ISA = ('Geo::Coder::Free::DB::MaxMind::admin');

1;
